from typing import NamedTuple

import jax
import jax.numpy as jnp
from jaxtyping import Float

from ..geostrophy import geostrophy
from ..utils import geometry, operators, sanitize


# =============================================================================
# Types
# =============================================================================

class CyclogeostrophySetup(NamedTuple):
    """
    Contains precomputed values needed for cyclogeostrophic calculations.

    Attributes
    ----------
    land_mask : Float[jax.Array, "y x"]
        Land mask where `True`/`1` indicates land
    ug_t : Float[jax.Array, "y x"]
        $u$ component of geostrophic velocity, on the T grid
    vg_t : Float[jax.Array, "y x"]
        $v$ component of geostrophic velocity, on the T grid
    dx_e_t : Float[jax.Array, "y x"]
        Eastward displacement associated with one step in the x-index direction, on the T grid
    dx_n_t : Float[jax.Array, "y x"]
        Northward displacement associated with one step in the x-index direction, on the T grid
    dy_e_t : Float[jax.Array, "y x"]
        Eastward displacement associated with one step in the y-index direction, on the T grid
    dy_n_t : Float[jax.Array, "y x"]
        Northward displacement associated with one step in the y-index direction, on the T grid
    J_t : Float[jax.Array, "y x"]
        Jacobian of the transformation from grid to geographic coordinates, on the T grid
    coriolis_factor_t : Float[jax.Array, "y x"]
        Coriolis factor, on the T grid
    """

    land_mask: Float[jax.Array, "y x"]
    ug_t: Float[jax.Array, "y x"]
    vg_t: Float[jax.Array, "y x"]
    dx_e_t: Float[jax.Array, "y x"]
    dx_n_t: Float[jax.Array, "y x"]
    dy_e_t: Float[jax.Array, "y x"]
    dy_n_t: Float[jax.Array, "y x"]
    J_t: Float[jax.Array, "y x"]
    coriolis_factor_t: Float[jax.Array, "y x"]


class CyclogeostrophyResult(NamedTuple):
    """
    Result of cyclogeostrophic velocity computation.

    This NamedTuple provides named access to results, avoiding positional unpacking errors.
    All fields except ``ucg`` and ``vcg`` are optional and depend on the
    ``return_*`` flags passed to the computation function.

    Attributes
    ----------
    ucg : Float[jax.Array, "y x"]
        $u$ component of cyclogeostrophic velocity, on the T grid
    vcg : Float[jax.Array, "y x"]
        $v$ component of cyclogeostrophic velocity, on the T grid
    ug : Float[jax.Array, "y x"] | None
        $u$ component of geostrophic velocity, on the T grid (if ``return_geos=True``)
    vg : Float[jax.Array, "y x"] | None
        $v$ component of geostrophic velocity, on the T grid (if ``return_geos=True``)
    losses : Float[jax.Array, "n_it"] | None
        Cyclogeostrophic imbalance over iterations (if ``return_losses=True``)
    """

    ucg: Float[jax.Array, "y x"]
    vcg: Float[jax.Array, "y x"]
    ug: Float[jax.Array, "y x"] | None = None
    vg: Float[jax.Array, "y x"] | None = None
    losses: Float[jax.Array, "n_it"] | None = None


# =============================================================================
# Setup and Result Assembly
# =============================================================================

def setup_cyclogeostrophy(
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    ssh_t: Float[jax.Array, "y x"] = None,
    ug_t: Float[jax.Array, "y x"] = None,
    vg_t: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None
) -> CyclogeostrophySetup:
    """
    Computes all preliminary values needed for cyclogeostrophic calculations.

    This includes: land mask, geostrophic velocities, U/V grids, spatial steps,
    Coriolis factors, and grid rotation angles on all grids.

    There are two modes of operation:

    1. **SSH mode**: Provide ``lat_t``, ``lon_t``, ``ssh_t`` (and optionally ``mask``).
       Geostrophic velocities will be computed from SSH

    2. **Geostrophic mode**: Provide ``lat_t``, ``lon_t``, ``ug_t``, ``vg_t``
       (and optionally ``land_mask``). Geostrophic velocities are provided on the T grid

    Parameters
    ----------
    lat_t : Float[jax.Array, "y x"]
        Latitudes of T grid.
    lon_t : Float[jax.Array, "y x"]
        Longitudes of T grid.
    ssh_t : Float[jax.Array, "y x"], optional
        SSH field on T grid. Required if geostrophic velocities are not provided.
    ug_t : Float[jax.Array, "y x"], optional
        U component of geostrophic velocity on T grid. If provided with ``vg_t``,
        bypasses SSH-based computation. Will be interpolated to U grid.
    vg_t : Float[jax.Array, "y x"], optional
        V component of geostrophic velocity on T grid. If provided with ``ug_t``,
        bypasses SSH-based computation. Will be interpolated to V grid.
    land_mask : Float[jax.Array, "y x"], optional
        Land mask where `1`/`True` is land. If None, inferred from ssh_t or ug_t nan values.

    Returns
    -------
    CyclogeostrophySetup
        Named tuple containing all precomputed values

    Raises
    ------
    ValueError
        If neither SSH nor geostrophic velocity inputs are provided.
    """
    # Check if geostrophic velocities are provided directly
    use_geos_directly = ug_t is not None and vg_t is not None

    if use_geos_directly:
        land_mask = sanitize.init_land_mask(ug_t, land_mask)
    else:
        # SSH-based computation
        if ssh_t is None:
            raise ValueError(
                "Either provide ssh_t to compute geostrophic velocities from SSH, "
                "or provide ug_t, vg_t directly on the T grid."
            )
        
        land_mask = sanitize.init_land_mask(ssh_t, land_mask)

        ug_t, vg_t = geostrophy(ssh_t, lat_t, lon_t, land_mask)

    dx_e, dx_n, dy_e, dy_n, J = geometry.grid_metrics(lat_t, lon_t)

    f = geometry.coriolis_factor(lat_t)

    return CyclogeostrophySetup(
        land_mask=land_mask,
        ug_t=ug_t,
        vg_t=vg_t,
        dx_e_t=dx_e,
        dx_n_t=dx_n,
        dy_e_t=dy_e,
        dy_n_t=dy_n,
        J_t=J,
        coriolis_factor_t=f,
    )


def assemble_result(
    ucg_t: Float[jax.Array, "y x"],
    vcg_t: Float[jax.Array, "y x"],
    setup: CyclogeostrophySetup,
    return_geos: bool = False,
    return_losses: bool = False,
    losses: Float[jax.Array, "n_it"] = None,
) -> CyclogeostrophyResult:
    """
    Assembles the final result, sanitizing outputs and including optional fields.

    Parameters
    ----------
    ucg_t : Float[jax.Array, "y x"]
        $u$ component of cyclogeostrophic velocity
    vcg_t : Float[jax.Array, "y x"]
        $v$ component of cyclogeostrophic velocity
    setup : CyclogeostrophySetup
        Precomputed setup values
    return_geos : bool, optional
        Include geostrophic velocities in result
    return_grids : bool, optional
        Include U/V grid coordinates in result
    return_losses : bool, optional
        Include losses in result
    losses : Float[jax.Array, "n_it"], optional
        Loss values from iterative methods

    Returns
    -------
    CyclogeostrophyResult
        Named tuple with computed velocities and optional fields
    """
    # Handle masked data (set land cells to NaN)
    ucg_t = sanitize.sanitize_data(ucg_t, jnp.nan, setup.land_mask)
    vcg_t = sanitize.sanitize_data(vcg_t, jnp.nan, setup.land_mask)

    return CyclogeostrophyResult(
        ucg=ucg_t,
        vcg=vcg_t,
        ug=setup.ug_t if return_geos else None,
        vg=setup.vg_t if return_geos else None,
        losses=losses if return_losses else None,
    )


# =============================================================================
# Public API Functions
# =============================================================================

def cyclogeostrophic_loss(
    ug: Float[jax.Array, "y x"],
    vg: Float[jax.Array, "y x"],
    ucg: Float[jax.Array, "y x"],
    vcg: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True
) -> Float[jax.Array, ""]:
    """
    Computes the cyclogeostrophic imbalance loss from a geostrophic and a cyclogeostrophic velocity field.

    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    If provided, the ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are expected to follow the NEMO convention.

    Parameters
    ----------
    ug : Float[jax.Array, "y x"]
        $u$ component of the geostrophic velocity field
    vg : Float[jax.Array, "y x"]
        $v$ component of the geostrophic velocity field
    ucg : Float[jax.Array, "y x"]
        $u$ component of the cyclogeostrophic velocity field
    vcg : Float[jax.Array, "y x"]
        $v$ component of the cyclogeostrophic velocity field
    lat_t : Float[jax.Array, "y x"], optional
        Latitudes of the T grid.
        
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lon_t : Float[jax.Array, "y x"], optional
        Longitudes of the T grid.
       
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lat_u : Float[jax.Array, "y x"], optional
        Latitudes of the U grid.
        
        Defaults to `None`
    lon_u : Float[jax.Array, "y x"], optional
        Longitudes of the U grid.
        
        Defaults to `None`
    lat_v : Float[jax.Array, "y x"], optional
        Latitudes of the V grid.
        
        Defaults to `None`
    lon_v : Float[jax.Array, "y x"], optional
        Longitudes of the V grid.
        
        Defaults to `None`
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the T grid 
        (this is important when manipulating staggered grids)
        
        Defaults to `True`
    Returns
    -------
    loss : Float[jax.Array, ""]
        Cyclogeostrophic imbalance loss
    """
    u_imbalance, v_imbalance = cyclogeostrophic_imbalance(
        ug, vg, ucg, vcg, lat_t, lon_t, lat_u, lon_u, lat_v, lon_v, land_mask, uv_on_t
    )

    return jnp.nansum(u_imbalance ** 2 + v_imbalance ** 2)


def cyclogeostrophic_imbalance(
    ug: Float[jax.Array, "y x"],
    vg: Float[jax.Array, "y x"],
    ucg: Float[jax.Array, "y x"],
    vcg: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    """
    Computes the cyclogeostrophic imbalance of a 2d velocity field.

    The velocity fields can be provided either on the U and V grids (``vel_on_uv=True``) or on the T grid (``vel_on_uv=False``).

    Parameters
    ----------
    ug : Float[jax.Array, "y x"]
        $u$ component of the geostrophic velocity field
    vg : Float[jax.Array, "y x"]
        $v$ component of the geostrophic velocity field
    ucg : Float[jax.Array, "y x"]
        $u$ component of the cyclogeostrophic velocity field
    vcg : Float[jax.Array, "y x"]
        $v$ component of the cyclogeostrophic velocity field
    lat_t : Float[jax.Array, "y x"], optional
        Latitudes of the T grid.
        
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lon_t : Float[jax.Array, "y x"], optional
        Longitudes of the T grid.
       
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lat_u : Float[jax.Array, "y x"], optional
        Latitudes of the U grid.
        
        Defaults to `None`
    lon_u : Float[jax.Array, "y x"], optional
        Longitudes of the U grid.
        
        Defaults to `None`
    lat_v : Float[jax.Array, "y x"], optional
        Latitudes of the V grid.
        
        Defaults to `None`
    lon_v : Float[jax.Array, "y x"], optional
        Longitudes of the V grid.
        
        Defaults to `None`
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the T grid 
        (this is important when manipulating staggered grids)
        
        Defaults to `True`

    Returns
    -------
    u_imbalance : Float[jax.Array, "y x"]
        $u$ component of the cyclogeostrophic imbalance, on the T grid
    v_imbalance : Float[jax.Array, "y x"]
        $v$ component of the cyclogeostrophic imbalance, on the T grid
    """
    if land_mask is None:
        land_mask = sanitize.init_land_mask(ug)

    if not uv_on_t:
        ug = operators.interpolation(ug, axis=1, padding="left", land_mask=land_mask)  # U(i), U(i+1) -> T(i+1)
        vg = operators.interpolation(vg, axis=0, padding="left", land_mask=land_mask)  # U(i), U(i+1) -> T(i+1)
        ucg = operators.interpolation(ucg, axis=1, padding="right", land_mask=land_mask)
        vcg = operators.interpolation(vcg, axis=0, padding="right", land_mask=land_mask)

    if lat_t is None or lon_t is None:
        if lat_u is not None and lon_u is not None:
            lat_t = operators.interpolation(lat_u, axis=1, padding="left", land_mask=land_mask)
            lon_t = operators.interpolation(lon_u, axis=1, padding="left", land_mask=land_mask)
        elif lat_v is not None and lon_v is not None:
            lat_t = operators.interpolation(lat_v, axis=0, padding="left", land_mask=land_mask)
            lon_t = operators.interpolation(lon_v, axis=0, padding="left", land_mask=land_mask)
        else:
            raise ValueError("Either lat_t and lon_t, or lat_u, lon_u, lat_v, and lon_v must be provided")
    
    # compute grid metrics once
    dx_e, dx_n, dy_e, dy_n, J = geometry.grid_metrics(lat_t, lon_t)
    f = geometry.coriolis_factor(lat_t)

    return _cyclogeostrophic_imbalance(
        ug, vg, ucg, vcg, dx_e, dx_n, dy_e, dy_n, J, f, land_mask
    )


# =============================================================================
# Internal Functions
# =============================================================================

def _cyclogeostrophic_loss(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    ucg_t: Float[jax.Array, "y x"],
    vcg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
) -> Float[jax.Array, ""]:
    u_imbalance, v_imbalance = _cyclogeostrophic_imbalance(
        ug_t, vg_t, ucg_t, vcg_t, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, coriolis_factor_t, land_mask
    )

    return jnp.nansum(u_imbalance ** 2 + v_imbalance ** 2)


def _cyclogeostrophic_imbalance(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    ucg_t: Float[jax.Array, "y x"],
    vcg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    u_adv_t, v_adv_t = _advection(ucg_t, vcg_t, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, land_mask)

    u_imbalance = ucg_t + v_adv_t / coriolis_factor_t - ug_t
    v_imbalance = vcg_t - u_adv_t / coriolis_factor_t - vg_t

    return u_imbalance, v_imbalance


def _advection(
    u_t: Float[jax.Array, "y x"],
    v_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    u_adv = _u_advection(u_t, v_t, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, land_mask)
    v_adv = _v_advection(u_t, v_t, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, land_mask)

    return u_adv, v_adv


def _u_advection(
    u_t: Float[jax.Array, "y x"],
    v_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
) -> Float[jax.Array, "y x"]:
    """
    Computes u * ∂u/∂x + v * ∂u/∂y
    """
    du_e_t, du_n_t = operators.horizontal_derivatives(
        u_t, dx_e=dx_e_t, dx_n=dx_n_t, dy_e=dy_e_t, dy_n=dy_n_t, J=J_t, land_mask=land_mask
    )

    u_adv = u_t * du_e_t + v_t * du_n_t

    return u_adv


def _v_advection(
    u_t: Float[jax.Array, "y x"],
    v_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
) -> Float[jax.Array, "y x"]:
    """
    Computes u * ∂v/∂x + v * ∂v/∂y
    """
    dv_e_t, dv_n_t = operators.horizontal_derivatives(
        v_t, dx_e=dx_e_t, dx_n=dx_n_t, dy_e=dy_e_t, dy_n=dy_n_t, J=J_t, land_mask=land_mask
    )

    v_adv = u_t * dv_e_t + v_t * dv_n_t

    return v_adv



from functools import partial

import jax
from jax import lax
import jax.numpy as jnp
from jaxtyping import Bool, Float

from ._core import (
    CyclogeostrophyResult, setup_cyclogeostrophy, assemble_result, _advection, _cyclogeostrophic_loss
)


def fixed_point(
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    ssh_t: Float[jax.Array, "y x"] = None,
    ug_t: Float[jax.Array, "y x"] = None,
    vg_t: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    return_geos: bool = False,
    return_losses: bool = False,
    n_it: int = 20,
    res_eps: float = 0.01
) -> CyclogeostrophyResult:
    """
    Computes the cyclogeostrophic Sea Surface Current (SSC) velocity field
    using the fixed-point method [Penven et al. (2014)](https://doi.org/10.1002/2013JC009528).

    There are two modes of operation:

    1. **SSH mode**: Provide ``lat_t``, ``lon_t``, ``ssh_t`` (and optionally ``mask``).
       Geostrophic velocities will be computed from SSH.

    2. **Geostrophic mode**: Provide ``lat_t``, ``lon_t``, ``ug_t``, ``vg_t``
       (and optionally ``land_mask``). Geostrophic velocities are provided on the T grid.

    Parameters
    ----------
    lat_t : Float[jax.Array, "y x"]
        Latitude of the T grid
    lon_t : Float[jax.Array, "y x"]
        Longitude of the T grid
    ssh_t : Float[jax.Array, "y x"], optional
        SSH field (on the T grid)

        Defaults to `None`, required if geostrophic velocities are not provided
    ug_t : Float[jax.Array, "y x"], optional
        U component of geostrophic velocity on T grid
        
        Defaults to `None`, required if ``ssh_t`` is not provided
    vg_t : Float[jax.Array, "y x"], optional
        V component of geostrophic velocity on T grid
        
        Defaults to `None`, required if ``ssh_t`` is not provided
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        If not provided, inferred from ``ssh_t`` or ``ug_t`` `nan` values

        Defaults to `None`
    return_geos : bool, optional
        If `True`, returns the geostrophic SSC velocity field in addition to the cyclogeostrophic one.

        Defaults to `False`
    return_losses : bool, optional
        If `True`, returns the losses (cyclogeostrophic imbalance) over iterations.

        Defaults to `False`
    n_it : int, optional
        Maximum number of iterations.

        Defaults to `20`
    res_eps : float, optional
        Residual tolerance of the iterative approach.
        When residuals are smaller, the iterative approach considers local convergence to cyclogeostrophy.

        Defaults to `0.01`

    Returns
    -------
    CyclogeostrophyResult
        Named tuple containing:
        - ``ucg``: $u$ component of cyclogeostrophic velocity, on the T grid
        - ``vcg``: $v$ component of cyclogeostrophic velocity, on the T grid
        - ``ug``, ``vg``: Geostrophic velocities (if ``return_geos=True``)
        - ``losses``: Cyclogeostrophic imbalance per iteration (if ``return_losses=True``)
    """
    setup = setup_cyclogeostrophy(
        lat_t, lon_t, ssh_t=ssh_t, ug_t=ug_t, vg_t=vg_t, land_mask=land_mask
    )

    ucg, vcg, losses = _fixed_point(
        setup.ug_t, setup.vg_t,
        setup.dx_e_t, setup.dx_n_t, setup.dy_e_t, setup.dy_n_t, setup.J_t,
        setup.coriolis_factor_t,
        setup.land_mask, n_it, res_eps, return_losses
    )

    return assemble_result(
        ucg, vcg, setup, return_geos=return_geos, return_losses=return_losses, losses=losses
    )


@partial(jax.jit, static_argnames=("n_it"))
def _fixed_point(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
    n_it: int,
    res_eps: float,
    return_losses: bool
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"], Float[jax.Array, "n_it"]]:
    # define step partial: freeze constant over iterations
    def step_fn(carry, _):
        return _fp_step(
            ug_t, vg_t,
            dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t,
            coriolis_factor_t,
            land_mask,
            res_eps, return_losses,
            *carry
        )

    # apply updates
    (ucg, vcg, _, _), losses = lax.scan(
        step_fn,
        (ug_t, vg_t, (1 - land_mask).astype(bool), jnp.maximum(jnp.abs(ug_t), jnp.abs(vg_t))),
        xs=None, length=n_it
    )

    return ucg, vcg, losses


def _fp_step(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
    res_eps: float,
    return_losses: bool,
    u_n: Float[jax.Array, "y x"],
    v_n: Float[jax.Array, "y x"],
    mask_update: Bool[jax.Array, "y x"],
    res_n: Float[jax.Array, "y x"]
) -> tuple[
    tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"], Float[jax.Array, "y x"], Float[jax.Array, "y x"]], 
    float
]:
    # compute loss
    loss = lax.cond(
        return_losses,
        lambda: _cyclogeostrophic_loss(
            ug_t, vg_t, u_n, v_n, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, coriolis_factor_t, land_mask
        ),
        lambda: jnp.nan
    )

    # next it
    u_adv, v_adv = _advection(u_n, v_n, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, land_mask)
    u_np1 = ug_t - jnp.nan_to_num(v_adv / coriolis_factor_t, copy=False, nan=0, posinf=0, neginf=0)
    v_np1 = vg_t + jnp.nan_to_num(u_adv / coriolis_factor_t, copy=False, nan=0, posinf=0, neginf=0)

    # compute dist to ucg and vcg
    res_np1 = jnp.abs(u_np1 - u_n) + jnp.abs(v_np1 - v_n)  # norm1

    # compute stopping criterion masks
    mask_not_div = jnp.where(res_np1 <= res_n, True, False)
    mask_not_conv = jnp.where(res_np1 >= res_eps, True, False)
  
    # update cyclogeostrophic velocities and residuals where it is not diverging
    mask_update &= mask_not_div
    u_n = jnp.where(mask_update, u_np1, u_n)
    v_n = jnp.where(mask_update, v_np1, v_n)
    res_n = jnp.where(mask_update, res_np1, res_n)

    # update stopping criterion mask where it has converged
    mask_update &= mask_not_conv

    return (u_n, v_n, mask_update, res_n), loss



from .version import __version__  # noqa

from .geostrophy import geostrophy
from .cyclogeostrophy import fixed_point, gradient_wind, minimization_based

__all__ = [
    "__version__",
    "geostrophy",
    "fixed_point",
    "gradient_wind",
    "minimization_based",
]



from typing import Literal

import jax
from jax import lax
import jax.numpy as jnp
from jaxtyping import Float


def interpolation(
    field: Float[jax.Array, "y x"],
    axis: Literal[0, 1],
    padding: Literal["left", "right"],
    land_mask: Float[jax.Array, "y x"] = None,
) -> Float[jax.Array, "y x"]:
    """
    Interpolates the values of a ``field`` along a given ``axis`` (`0` for `lat`/`y`, `1` for `lon`/`x`),
    applying ``padding`` to the `left` (i.e. `West` if ``axis=1``, `South` if ``axis=0``) or
    to the `right` (i.e. `East` if ``axis=1``, `North` if ``axis=0``) of the domain.

    An open boundary condition is applied:

    - At domain edges: the interpolated value equals the nearest interior value
    - At land/NaN boundaries: if one of the two values is NaN, the valid value is used;
      if both are NaN, the result is NaN

    Parameters
    ----------
    field : Float[jax.Array, "y x"]
        Field to interpolate
    axis : Literal[0, 1]
        Axis along which interpolation is performed
    padding : Literal["left", "right"]
        Padding direction.
        For example, following NEMO convention,
        interpolating from U to T points requires a `left` padding
        (the midpoint between $U_i$ and $U_{i+1}$ corresponds to $T_{i+1}$),
        and interpolating from T to U points a `right` padding
        (the midpoint between $T_i$ and $T_{i+1}$ corresponds to $U_i$)
    land_mask : Float[jax.Array, "y x"], optional
        Mask indicating the land domain where extrapolation should be applied.
        `False`/`0` indicates ocean cells, `True`/`1` indicates land cells.

        Defaults to `None`, 
        in which case no land masking is applied and extrapolation is performed across the entire domain

    Returns
    -------
    field : Float[jax.Array, "y x"]
        Interpolated field
    """
    f = jnp.moveaxis(field, axis, -1)

    left = f[:, :-1]
    right = f[:, 1:]

    left_valid = ~jnp.isnan(left)
    right_valid = ~jnp.isnan(right)

    # Open boundary condition for NaN values:
    # - Both valid: average
    # - One valid: use the valid one
    # - Both NaN: NaN
    mid = jnp.where(
        left_valid & right_valid,
        (left + right) * 0.5,
        jnp.where(left_valid, left, jnp.where(right_valid, right, jnp.nan))
    )

    # Pad at the domain boundary with edge value (open boundary condition)
    mid = lax.cond(
        padding == "left",
        lambda: jnp.pad(mid, ((0, 0), (1, 0)), mode='edge'),
        lambda: jnp.pad(mid, ((0, 0), (0, 1)), mode='edge')
    )

    mid = jnp.moveaxis(mid, -1, axis)

    if land_mask is not None:
        mid = jnp.where(land_mask, jnp.nan, mid)

    return mid


def derivative(
    field: Float[jax.Array, "y x"],
    axis: Literal[0, 1],
    padding: Literal["left", "right"],
    land_mask: Float[jax.Array, "y x"] = None,
) -> Float[jax.Array, "y x"]:
    """
    Differentiates a ``field``, using finite differences, along a given ``axis`` (`0` for `lat`/`y`, `1` for `lon`/'x'),
    applying ``padding`` to the `left` (i.e. `West` if ``axis=1``, `South` if ``axis=0``) or
    to the `right` (i.e. `East` if ``axis=1``, `North` if ``axis=0``) of the domain.

    An open boundary condition is applied (zero second derivative):

    - At domain edges: the boundary derivative equals the nearest interior derivative
    - At land/NaN boundaries: if one of the two values is NaN, the derivative is filled
      with the immediate neighbor derivative; if both neighbors are NaN, the result is NaN

    This is appropriate for domains with sharp physical boundaries (e.g., SWOT swaths)
    where the signal continues smoothly beyond the observation edge.

    Parameters
    ----------
    field : Float[jax.Array, "y x"]
        Field to differentiate
    axis : Literal[0, 1]
        Axis along which interpolation is performed
    padding : Literal["left", "right"]
        Padding direction.
        For example, following NEMO convention,
        interpolating from U to T points requires a `left` padding
        (the midpoint between $U_i$ and $U_{i+1}$ corresponds to $T_{i+1}$),
        and interpolating from T to U points a `right` padding
        (the midpoint between $T_i$ and $T_{i+1}$ corresponds to $U_i$)
    land_mask : Float[jax.Array, "y x"], optional
        Mask indicating the land domain where extrapolation should be applied.
        `False`/`0` indicates ocean cells, `True`/`1` indicates land cells.

        Defaults to `None`, 
        in which case no land masking is applied and extrapolation is performed across the entire domain

    Returns
    -------
    df : Float[jax.Array, "y x"]
        Differentiated field
    """
    f = jnp.moveaxis(field, axis, -1)

    df = jnp.diff(f, axis=-1)

    # Open boundary condition for NaN values:
    # Fill NaN derivatives with immediate neighbor
    left_neighbor = jnp.roll(df, 1, axis=-1).at[..., 0].set(jnp.nan)
    right_neighbor = jnp.roll(df, -1, axis=-1).at[..., -1].set(jnp.nan)

    df = jnp.where(
        ~jnp.isnan(df),
        df,
        jnp.where(
            ~jnp.isnan(left_neighbor), left_neighbor,
            jnp.where(~jnp.isnan(right_neighbor), right_neighbor, jnp.nan)
        )
    )

    # Open boundary condition at domain edges: ∂²f/∂x² = 0
    # The boundary derivative equals the nearest interior derivative
    df = lax.cond(
        padding == "left",
        lambda: jnp.concatenate([df[..., 0:1], df], axis=-1),
        lambda: jnp.concatenate([df, df[..., -1:]], axis=-1)
    )

    df = jnp.moveaxis(df, -1, axis)

    if land_mask is not None:
        df = jnp.where(land_mask, jnp.nan, df)

    return df


def horizontal_derivatives(
    field: Float[jax.Array, "y x"],
    lat: Float[jax.Array, "y x"] = None,
    lon: Float[jax.Array, "y x"] = None,
    dx_e: Float[jax.Array, "y x"] = None,
    dx_n: Float[jax.Array, "y x"] = None,
    dy_e: Float[jax.Array, "y x"] = None,
    dy_n: Float[jax.Array, "y x"] = None,
    J: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    """
    Computes the horizontal derivatives of a ``field`` defined on a curvilinear or rectilinear grid, 
    using finite differences and applying an open boundary condition at domain edges and land/NaN boundaries.

    Horizontal derivatives are returned in the same grid as the input field.

    Parameters
    ----------
    field : Float[jax.Array, "y x"]
        Field for which to compute gradients
    lat : Float[jax.Array, "y x"], optional
        Latitude grid corresponding to the field

        Defaults to `None`, in which case ``dx_e``, ``dx_n``, ``dy_e``, ``dy_n`` and ``J`` must be provided
    lon : Float[jax.Array, "y x"], optional
        Longitude grid corresponding to the field

        Defaults to `None`, in which case ``dx_e``, ``dx_n``, ``dy_e``, ``dy_n`` and ``J`` must be provided
    dx_e : Float[jax.Array, "y x"], optional
        Grid spacing in the eastward direction (i.e. along axis=1)

        Defaults to `None`, in which case ``lat`` and ``lon`` must be provided
    dx_n : Float[jax.Array, "y x"], optional
        Grid spacing in the northward direction (i.e. along axis=0)

        Defaults to `None`, in which case ``lat`` and ``lon`` must be provided
    dy_e : Float[jax.Array, "y x"], optional
        Grid spacing in the eastward direction (i.e. along axis=1)

        Defaults to `None`, in which case ``lat`` and ``lon`` must be provided
    dy_n : Float[jax.Array, "y x"], optional
        Grid spacing in the northward direction (i.e. along axis=0)

        Defaults to `None`, in which case ``lat`` and ``lon`` must be provided
    J : Float[jax.Array, "y x"], optional
        Jacobian of the transformation from grid to geographic coordinates

        Defaults to `None`, in which case ``lat`` and ``lon`` must be provided
    land_mask : Float[jax.Array, "y x"], optional
        Mask indicating the land domain where extrapolation should be applied.
        `False`/`0` indicates ocean cells, `True`/`1` indicates land cells.

        Defaults to `None`, 
        in which case no land masking is applied and extrapolation is performed across the entire domain
    
    Returns
    -------
    df_e : Float[jax.Array, "y x"]
        Eastward derivative of the field, on the same grid as the input field
    df_n : Float[jax.Array, "y x"]
        Northward derivative of the field, on the same grid as the input field
    """
    if dx_e is None or dx_n is None or dy_e is None or dy_n is None or J is None:
        if lat is None or lon is None:
            raise ValueError("Either lat/lon or dx_e/dx_n/dy_e/dy_n/J must be provided")
        from .geometry import grid_metrics
        dx_e, dx_n, dy_e, dy_n, J = grid_metrics(lat, lon)

    # compute derivatives in grid coordinates
    df_x_s1 = derivative(field, axis=1, padding="right", land_mask=land_mask)
    df_y_s0 = derivative(field, axis=0, padding="right", land_mask=land_mask)

    # because of staggered grid, we need to interpolate
    df_x = interpolation(df_x_s1, axis=1, padding="left", land_mask=land_mask)
    df_y = interpolation(df_y_s0, axis=0, padding="left", land_mask=land_mask)

    # transform derivatives to geographic coordinates
    df_e = (df_x * dy_n - df_y * dy_e) / J
    df_n = (df_y * dx_e - df_x * dx_n) / J

    return df_e, df_n



import jax
import jax.numpy as jnp
from jaxtyping import Float

from .utils import geometry, operators, sanitize


# =============================================================================
# Geostrophy
# =============================================================================

def geostrophy(
    ssh_t: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"] = None
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    """
    Computes the geostrophic Sea Surface Current (SSC) velocity field from a Sea Surface Height (SSH) field.

    Parameters
    ----------
    ssh_t : Float[jax.Array, "y x"]
        SSH field (on the T grid)
    lat_t : Float[jax.Array, "y x"]
        Latitudes of the T grid
    lon_t : Float[jax.Array, "y x"]
        Longitudes of the T grid
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land).

        Defaults to `None`, in which case inferred from `ssh_t` `nan` values

    Returns
    -------
    ug_t : Float[jax.Array, "y x"]
        $u$ component of the geostrophic velocity field, on the T grid
    vg_t : Float[jax.Array, "y x"]
        $v$ component of the geostrophic velocity field, on the T grid
    """
    # Make sure the mask is initialized
    land_mask = sanitize.init_land_mask(ssh_t, land_mask)

    # Handle spurious and masked data
    ssh_t = sanitize.sanitize_data(ssh_t, jnp.nan, land_mask)

    ug_t, vg_t = _geostrophy(ssh_t, lat_t, lon_t, land_mask)

    # Handle masked data (set land cells to NaN)
    ug_t = sanitize.sanitize_data(ug_t, jnp.nan, land_mask)
    vg_t = sanitize.sanitize_data(vg_t, jnp.nan, land_mask)

    return ug_t, vg_t


@jax.jit
def _geostrophy(
    ssh_t: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"]
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    """
    Internal JIT-compiled geostrophy computation with curvilinear grid support.
    """
    deta_e_t, deta_n_t = operators.horizontal_derivatives(ssh_t, lat=lat_t, lon=lon_t, land_mask=land_mask)

    f_t = geometry.coriolis_factor(lat_t)

    # Computing the geostrophic velocities
    # u = -g/f * dη/dn  (eastward velocity)
    # v =  g/f * dη/de  (northward velocity)
    ug_t = -geometry.GRAVITY * deta_n_t / f_t
    vg_t = geometry.GRAVITY * deta_e_t / f_t

    return ug_t, vg_t



from functools import partial

import jax
from jax import lax
import jax.numpy as jnp
from jaxtyping import Float
import optax

from ._core import CyclogeostrophyResult, setup_cyclogeostrophy, assemble_result, _cyclogeostrophic_loss


def minimization_based(
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    ssh_t: Float[jax.Array, "y x"] = None,
    ug_t: Float[jax.Array, "y x"] = None,
    vg_t: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    return_geos: bool = False,
    return_losses: bool = False,
    n_it: int = 2000,
    optim: optax.GradientTransformation | str = "sgd",
    optim_kwargs: dict = None
) -> CyclogeostrophyResult:
    """
    Computes the cyclogeostrophic Sea Surface Current (SSC) velocity field
    using our minimization-based method.

    There are two modes of operation:

    1. **SSH mode**: Provide ``lat_t``, ``lon_t``, ``ssh_t`` (and optionally ``mask``).
       Geostrophic velocities will be computed from SSH.

    2. **Geostrophic mode**: Provide ``lat_t``, ``lon_t``, ``ug_t``, ``vg_t``
       (and optionally ``mask``). Geostrophic velocities are provided on the T grid
       and will be interpolated to U/V grids internally.

    Parameters
    ----------
    lat_t : Float[jax.Array, "y x"]
        Latitude of the T grid.
    lon_t : Float[jax.Array, "y x"]
        Longitude of the T grid.
    ssh_t : Float[jax.Array, "y x"], optional
        SSH field (on the T grid). Required if geostrophic velocities are not provided.
    ug_t : Float[jax.Array, "y x"], optional
        U component of geostrophic velocity on T grid. If provided with ``vg_t``,
        bypasses SSH-based computation. Will be interpolated to U grid.
    vg_t : Float[jax.Array, "y x"], optional
        V component of geostrophic velocity on T grid. If provided with ``ug_t``,
        bypasses SSH-based computation. Will be interpolated to V grid.
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        If not provided, inferred from ``ssh_t`` or ``ug_t`` `nan` values

        Defaults to `None`
    return_geos : bool, optional
        If `True`, returns the geostrophic SSC velocity field in addition to the cyclogeostrophic one.

        Defaults to `False`
    return_grids : bool, optional
        If `True`, returns the U and V grids.

        Defaults to `True`
    return_losses : bool, optional
        If `True`, returns the losses (cyclogeostrophic imbalance) over iterations.

        Defaults to `False`
    n_it : int, optional
        Maximum number of iterations.

        Defaults to `2000`
    optim : Union[optax.GradientTransformation, str], optional
        Optimizer to use.
        Can be an ``optax.GradientTransformation`` optimizer, or a ``string`` referring to such an optimizer.

        Defaults to `sgd`
    optim_kwargs : dict, optional
        Optimizer arguments (such as learning rate, etc...).

        If `None`, only the learning rate is enforced to `0.005`

        Defaults to `None`

    Returns
    -------
    CyclogeostrophyResult
        Named tuple containing:
        - ``ucg``: $u$ component of cyclogeostrophic velocity, on the T grid
        - ``vcg``: $v$ component of cyclogeostrophic velocity, on the T grid
        - ``ug``, ``vg``: Geostrophic velocities (if ``return_geos=True``)
        - ``losses``: Cyclogeostrophic imbalance per iteration (if ``return_losses=True``)
    """
    setup = setup_cyclogeostrophy(
        lat_t, lon_t, ssh_t=ssh_t, ug_t=ug_t, vg_t=vg_t, land_mask=land_mask
    )

    if isinstance(optim, str):
        if optim_kwargs is None:
            optim_kwargs = {"learning_rate": 0.005}
        optim = getattr(optax, optim)(**optim_kwargs)
    elif not isinstance(optim, optax.GradientTransformation):
        raise TypeError(
            "optim should be an optax.GradientTransformation optimizer, or a string referring to such an optimizer."
        )
    
    ucg, vcg, losses = _minimization_based(
        setup.ug_t, setup.vg_t,
        setup.dx_e_t, setup.dx_n_t, setup.dy_e_t, setup.dy_n_t, setup.J_t,
        setup.coriolis_factor_t,
        setup.land_mask, n_it, optim
    )

    return assemble_result(
        ucg, vcg, setup, return_geos=return_geos, return_losses=return_losses, losses=losses
    )


@partial(jax.jit, static_argnames=("n_it", "optim"))
def _minimization_based(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"],
    n_it: int,
    optim: optax.GradientTransformation
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"], Float[jax.Array, "n_it"]]:
    def loss_fn(args):
        ucg, vcg = args
        return _cyclogeostrophic_loss(
            ug_t, vg_t, ucg, vcg, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, coriolis_factor_t, land_mask
        )
    
    def step_fn(carry, _):
        params = carry[:-1]
        opt_state = carry[-1]
    
        loss, grads = jax.value_and_grad(loss_fn)(params)
        grads = tuple(map(lambda x: jnp.nan_to_num(x, copy=False, nan=0, posinf=0, neginf=0), grads))

        updates, opt_state = optim.update(grads, opt_state, params)
        params = optax.apply_updates(params, updates)

        return params + (opt_state,), loss
    
    carry, losses = lax.scan(step_fn, (ug_t, vg_t, optim.init((ug_t, vg_t))), xs=None, length=n_it)
    
    return *carry[:-1], losses



import jax.numpy as jnp

from jaxparrow import fixed_point, geostrophy, gradient_wind, minimization_based

import gaussian_eddy


class TestVelocities:
    lat, lon, ssh, ug, vg, ucg, vcg, land_mask = gaussian_eddy.simulate_gaussian_eddy(R0=50e3, eta0=-.2)

    def test_geostrophy(self):
        ug_est, vg_est = geostrophy(ssh_t=self.ssh, lat_t=self.lat, lon_t=self.lon, land_mask=self.land_mask)
        geos_rmse = self.compute_rmse(self.ug, self.vg, ug_est, vg_est)  # around 0.0004623004
        print("Geos RMSE:", geos_rmse)
        assert geos_rmse < .0005

    def test_cyclogeostrophy_fixed_point(self):
        res_fp = fixed_point(
            lat_t=self.lat, lon_t=self.lon, ug_t=self.ug, vg_t=self.vg, land_mask=self.land_mask
        )

        ucg_est, vcg_est = res_fp.ucg, res_fp.vcg
        fp_rmse = self.compute_rmse(self.ucg, self.vcg, ucg_est, vcg_est)  # around 0.00035087694
        print("FP RMSE:", fp_rmse)

        assert fp_rmse < .0004

    def test_cyclogeostrophy_gradient_wind(self):
        res_gw = gradient_wind(
            lat_t=self.lat, lon_t=self.lon, ug_t=self.ug, vg_t=self.vg, land_mask=self.land_mask
        )

        ucg_est, vcg_est = res_gw.ucg, res_gw.vcg
        gw_rmse = self.compute_rmse(self.ucg, self.vcg, ucg_est, vcg_est)  # around 1.2627806e-05
        print("GW RMSE:", gw_rmse)

        assert gw_rmse < 1.3e-5

    def test_cyclogeostrophy_minimization(self):
        res_mb = minimization_based(
            lat_t=self.lat, lon_t=self.lon, ug_t=self.ug, vg_t=self.vg, land_mask=self.land_mask
        )

        ucg_est, vcg_est = res_mb.ucg, res_mb.vcg
        mb_rmse = self.compute_rmse(self.ucg, self.vcg, ucg_est, vcg_est)  # around 6.180092e-05
        print("MB RMSE:", mb_rmse)

        assert mb_rmse < 6.2e-5

    @staticmethod
    def compute_rmse(u, v, u_est, v_est) -> float:
        x = jnp.stack([u, v], axis=-1)
        x_est = jnp.stack([u_est, v_est], axis=-1)
        return jnp.sqrt(jnp.nanmean((x - x_est)**2))



import jax
import jax.numpy as jnp
from jaxtyping import Float


def sanitize_data(
    arr: Float[jax.Array, "y x"],
    fill_value: float,
    land_mask: Float[jax.Array, "y x"]
) -> Float[jax.Array, "y x"]:
    """
    Sanitizes data by replacing `nan` with ``fill_value`` and applying ``fill_value`` to the masked area.

    Parameters
    ----------
    arr : Float[jax.Array, "y x"]
        Array to sanitize
    fill_value : float
        Value to replace `nan` values and masked area with
    land_mask :  Float[jax.Array, "y x"]
        Mask to apply, `1` or `True` for masked

    Returns
    -------
    arr : Float[jax.Array, "y x"]
        Sanitized array
    """
    arr = jnp.nan_to_num(arr, copy=False, nan=fill_value, posinf=fill_value, neginf=fill_value)
    arr = jnp.where(land_mask, fill_value, arr)
    return arr


def init_land_mask(
    field: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"] = None
) -> Float[jax.Array, "y x"]:
    """
    If ``land_mask is None``, initializes it from the `nan` values of ``field``.
    If ``land_mask is not None``, simply returns it.

    Parameters
    ----------
    field : Float[jax.Array, "y x"]
        Field used to initialize the mask (if needed)
    land_mask :  Float[jax.Array, "y x"], optional
        Mask to initialized (if `None`).

        Defaults to `None`

    Returns
    -------
    land_mask : Float[jax.Array, "y x"]
        Initialized (if needed) land mask
    """
    if land_mask is None:
        land_mask = ~jnp.isfinite(field)
    return land_mask



import jax
import jax.numpy as jnp
from jaxtyping import Float

from ._core import (
    CyclogeostrophyResult, setup_cyclogeostrophy, assemble_result
)
from ..utils import kinematics


def gradient_wind(
    lat_t: Float[jax.Array, "y x"],
    lon_t: Float[jax.Array, "y x"],
    ssh_t: Float[jax.Array, "y x"] = None,
    ug_t: Float[jax.Array, "y x"] = None,
    vg_t: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    return_geos: bool = False
) -> CyclogeostrophyResult:
    """
    Computes the cyclogeostrophic Sea Surface Current (SSC) velocity field
    using the gradient wind approximation.

    There are two modes of operation:

    1. **SSH mode**: Provide ``lat_t``, ``lon_t``, ``ssh_t`` (and optionally ``mask``).
       Geostrophic velocities will be computed from SSH.

    2. **Geostrophic mode**: Provide ``lat_t``, ``lon_t``, ``ug_t``, ``vg_t``
       (and optionally ``land_mask``). Geostrophic velocities are provided on the T grid.

    Parameters
    ----------
    lat_t : Float[jax.Array, "y x"]
        Latitude of the T grid
    lon_t : Float[jax.Array, "y x"]
        Longitude of the T grid
    ssh_t : Float[jax.Array, "y x"], optional
        SSH field (on the T grid)

        Defaults to `None`, required if geostrophic velocities are not provided
    ug_t : Float[jax.Array, "y x"], optional
        U component of geostrophic velocity on T grid
        
        Defaults to `None`, required if ``ssh_t`` is not provided
    vg_t : Float[jax.Array, "y x"], optional
        V component of geostrophic velocity on T grid
        
        Defaults to `None`, required if ``ssh_t`` is not provided
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        If not provided, inferred from ``ssh_t`` or ``ug_t`` `nan` values

        Defaults to `None`
    return_geos : bool, optional
        If `True`, returns the geostrophic SSC velocity field in addition to the cyclogeostrophic one.

        Defaults to `False`

    Returns
    -------
    CyclogeostrophyResult
        Named tuple containing:
        - ``ucg``: $u$ component of cyclogeostrophic velocity, on the T grid
        - ``vcg``: $v$ component of cyclogeostrophic velocity, on the T grid
        - ``ug``, ``vg``: Geostrophic velocities (if ``return_geos=True``)
    """
    setup = setup_cyclogeostrophy(
        lat_t, lon_t, ssh_t=ssh_t, ug_t=ug_t, vg_t=vg_t, land_mask=land_mask
    )

    ucg, vcg = _gradient_wind(
        setup.ug_t, setup.vg_t,
        setup.dx_e_t, setup.dx_n_t, setup.dy_e_t, setup.dy_n_t, setup.J_t,
        setup.coriolis_factor_t, 
        setup.land_mask
    )

    return assemble_result(ucg, vcg, setup, return_geos=return_geos)


@jax.jit
def _gradient_wind(
    ug_t: Float[jax.Array, "y x"],
    vg_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    coriolis_factor_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"]
) -> tuple[Float[jax.Array, "y x"], Float[jax.Array, "y x"]]:
    R = kinematics._radius_of_curvature(ug_t, vg_t, dx_e_t, dx_n_t, dy_e_t, dy_n_t, J_t, land_mask)

    V_g = kinematics.magnitude(ug_t, vg_t, land_mask, uv_on_t=True)
    V_gr = 2 * V_g / (1 + jnp.sqrt(1 + 4 * V_g / (coriolis_factor_t * R)))

    ratio = V_gr / V_g

    ucg = ratio * ug_t
    vcg = ratio * vg_t

    return ucg, vcg



import jax
import jax.numpy as jnp

from jaxparrow.utils import geometry


def simulate_gaussian_eddy(
    R0: float,
    eta0: float,
) -> tuple[
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array, 
    jax.Array
]:
    # in the Alboran Sea
    latlon0 = (35.92744, -4.03238)
    # grid resolution (in °)
    dlatlon = 0.01

    # domain size (in °)
    hw_deg = 2 * R0 / 111e3
    # create lat lon grids
    lat_1d = latlon0[0] + jnp.arange(-hw_deg, hw_deg + dlatlon, dlatlon)
    lon_1d = latlon0[1] + jnp.arange(-hw_deg, hw_deg + dlatlon, dlatlon)
    lon, lat = jnp.meshgrid(lon_1d, lat_1d)

    # grid distance to latlon0 (in m)
    R = haversine_distance(lat, latlon0[0], lon, latlon0[1])
    # distace along x and y axes (in m)
    X = haversine_distance(lat, lat, lon, latlon0[1])
    Y = haversine_distance(lat, latlon0[0], lon, lon)
    # correct signs
    X = jnp.where(lon < latlon0[1], -X, X)
    Y = jnp.where(lat < latlon0[0], -Y, Y)

    coriolis_factor = jnp.ones_like(R) * geometry.coriolis_factor(lat)

    ssh = simulate_gaussian_ssh(R0, eta0, R)
    ug, vg = simulate_gaussian_geos(R0, X, Y, ssh, coriolis_factor)
    ucg, vcg = simulate_gaussian_cyclo(R0, X, Y, R, ssh, coriolis_factor)

    land_mask = jax.numpy.full_like(ssh, 0, dtype=bool)  # no land in square oceans

    return lat, lon, ssh, ug, vg, ucg, vcg, land_mask


def haversine_distance(lat, lat0, lon, lon0):
     # convert to radians
    lat_rad = jnp.radians(lat)
    lat0_rad = jnp.radians(lat0)

    # difference in radians; normalize lon diff to [-180, 180] before radians to handle dateline
    dlon = lon - lon0
    dlon = (dlon + 180.0) % 360.0 - 180.0   # now in [-180,180]
    dlon_rad = jnp.radians(dlon)

    dlat_rad = jnp.radians(lat - lat0)

    # haversine distance
    a = jnp.sin(dlat_rad / 2.0) ** 2 + jnp.cos(lat0_rad) * jnp.cos(lat_rad) * (jnp.sin(dlon_rad / 2.0) ** 2)
    c = 2.0 * jnp.arctan2(jnp.sqrt(a), jnp.sqrt(1.0 - a))
    d = geometry.EARTH_RADIUS * c

    return d


def simulate_gaussian_ssh(R0: float, eta0: float, r: jax.Array) -> jax.Array:
    return eta0 * jnp.exp(-(r / R0)**2)


def simulate_gaussian_geos(
    R0: float,
    X: jax.Array,
    Y: jax.Array,
    ssh: jax.Array,
    coriolis_factor: jax.Array
) -> tuple[jax.Array, jax.Array]:
    def f():
        return 2 * geometry.GRAVITY * ssh / (coriolis_factor * R0 ** 2)
    ug = Y * f()
    vg = -X * f()
    return ug, vg


def simulate_gaussian_cyclo(
    R0: float,
    X: jax.Array,
    Y: jax.Array,
    r: jax.Array,
    ssh: jax.Array,
    coriolis_factor: jax.Array
) -> tuple[jax.Array, jax.Array]:
    azim_geos = -(2 * geometry.GRAVITY * r * ssh / (coriolis_factor * R0 ** 2))
    azim_cyclo = 2 * azim_geos / (1 + jnp.sqrt(1 + 4 * azim_geos / (coriolis_factor * r)))
    ucg = -azim_cyclo * Y / r
    vcg = azim_cyclo * X / r
    return ucg, vcg



"""Utility modules for jaxparrow computations."""

from . import geometry
from . import kinematics
from . import operators
from . import sanitize

__all__ = ["geometry", "kinematics", "operators", "sanitize"]
 


import jax
import jax.numpy as jnp
from jaxtyping import Float

from .operators import interpolation


#: Approximate earth angular speed
EARTH_ANG_SPEED = 7.292115e-5
#: Approximate earth radius
EARTH_RADIUS = 6370e3
#: Approximate gravity
GRAVITY = 9.81


def grid_metrics(
    lat: Float[jax.Array, "y x"], lon: Float[jax.Array, "y x"]
) -> tuple[
    Float[jax.Array, "y x"], 
    Float[jax.Array, "y x"], 
    Float[jax.Array, "y x"], 
    Float[jax.Array, "y x"], 
    Float[jax.Array, "y x"]
]:
    """
    Computes the physical displacement vectors and the Jacobian associated with one grid-index step, 
    used to transform derivatives from grid coordinates to geographic coordinates.

    It makes use of the distance-on-a-sphere formula with Taylor expansion approximations of `cos` and `arccos`
    functions to avoid truncation issues.

    Parameters
    ----------
    lat : Float[jax.Array, "y x"]
        Latitude grid
    lon : Float[jax.Array, "y x"]
        Longitude grid

    Returns
    -------
    dx_e : Float[jax.Array, "y x"]
        Eastward displacement associated with one step in the x-index direction
    dx_n : Float[jax.Array, "y x"]
        Northward displacement associated with one step in the x-index direction
    dy_e : Float[jax.Array, "y x"]
        Eastward displacement associated with one step in the y-index direction
    dy_n : Float[jax.Array, "y x"]
        Northward displacement associated with one step in the y-index direction
    J : Float[jax.Array, "y x"]
        Jacobian of the transformation from grid to geographic coordinates
    """
    def displacement_components(lat1, lat2, lon1, lon2):
        # convert to radians
        lat1_rad = jnp.radians(lat1)
        lat2_rad = jnp.radians(lat2)

        # difference in radians; normalize lon diff to [-180, 180] before radians to handle dateline
        dlon = lon2 - lon1
        dlon = (dlon + 180.0) % 360.0 - 180.0   # now in [-180,180]
        dlon_rad = jnp.radians(dlon)

        dlat_rad = jnp.radians(lat2 - lat1)

        # haversine distance
        a = jnp.sin(dlat_rad / 2.0) ** 2 + jnp.cos(lat1_rad) * jnp.cos(lat2_rad) * (jnp.sin(dlon_rad / 2.0) ** 2)
        c = 2.0 * jnp.arctan2(jnp.sqrt(a), jnp.sqrt(1.0 - a))
        d = EARTH_RADIUS * c

        # bearing from point 1 to point 2
        theta = jnp.arctan2(
            jnp.sin(dlon_rad) * jnp.cos(lat2_rad),
            jnp.cos(lat1_rad) * jnp.sin(lat2_rad) - jnp.sin(lat1_rad) * jnp.cos(lat2_rad) * jnp.cos(dlon_rad)
        )

        de = d * jnp.sin(theta)
        dn = d * jnp.cos(theta)

        return de, dn

    # physical displacements
    dy_e, dy_n = displacement_components(lat[:-1, :], lat[1:, :], lon[:-1, :], lon[1:, :])
    dx_e, dx_n = displacement_components(lat[:, :-1], lat[:, 1:], lon[:, :-1], lon[:, 1:])

    dx_e = jnp.pad(dx_e, ((0, 0), (0, 1)), mode="edge")
    dx_n = jnp.pad(dx_n, ((0, 0), (0, 1)), mode="edge")
    dy_e = jnp.pad(dy_e, ((0, 1), (0, 0)), mode="edge")
    dy_n = jnp.pad(dy_n, ((0, 1), (0, 0)), mode="edge")

    # jacobian
    J = dx_e * dy_n - dx_n * dy_e

    return dx_e, dx_n, dy_e, dy_n, J


def coriolis_factor(lat: Float[jax.Array, "y x"]) -> Float[jax.Array, "y x"]:
    """
    Computes the Coriolis factor from a latitude grid.

    Parameters
    ----------
    lat : Float[jax.Array, "y x"]
        Latitudes grid

    Returns
    -------
    cf : Float[jax.Array, "y x"]
        Coriolis factor grid
    """
    return 2 * EARTH_ANG_SPEED * jnp.sin((jnp.radians(lat)))


def compute_uv_grids(
    lat_t: Float[jax.Array, "lat lon"],
    lon_t: Float[jax.Array, "lat lon"]
) -> tuple[
    Float[jax.Array, "lat lon"], Float[jax.Array, "lat lon"], Float[jax.Array, "lat lon"], Float[jax.Array, "lat lon"]
]:
    """
    Computes the U and V grids associated to a T grid following NEMO convention.

    Parameters
    ----------
    lat_t : Float[jax.Array, "lat lon"]
        Latitudes of the T grid
    lon_t : Float[jax.Array, "lat lon"]
        Longitudes of the T grid

    Returns
    -------
    lat_u : Float[jax.Array, "lat lon"]
        Latitudes of the U grid
    lon_u : Float[jax.Array, "lat lon"]
        Longitudes of the U grid
    lat_v : Float[jax.Array, "lat lon"]
        Latitudes of the V grid
    lon_v : Float[jax.Array, "lat lon"]
        Longitudes of the V grid
    """
    lat_u = interpolation(lat_t, axis=1, padding="right")
    lat_u = lat_u.at[:, -1].set(2 * lat_t[:, -1] - lat_t[:, -2])
    lon_u = interpolation(lon_t, axis=1, padding="right")
    lon_u = lon_u.at[:, -1].set(2 * lon_t[:, -1] - lon_t[:, -2])

    lat_v = interpolation(lat_t, axis=0, padding="right")
    lat_v = lat_v.at[-1, :].set(2 * lat_t[-1, :] - lat_t[-2, :])
    lon_v = interpolation(lon_t, axis=0, padding="right")
    lon_v = lon_v.at[-1, :].set(2 * lon_t[-1, :] - lon_t[-2, :])

    return lat_u, lon_u, lat_v, lon_v



from ._core import (
    CyclogeostrophyResult,
    CyclogeostrophySetup,
    cyclogeostrophic_loss,
    cyclogeostrophic_imbalance,
)
from ._fixed_point import fixed_point
from ._gradient_wind import gradient_wind
from ._minimization_based import minimization_based

__all__ = [
    "fixed_point",
    "gradient_wind",
    "minimization_based",
    "cyclogeostrophic_loss",
    "cyclogeostrophic_imbalance",
    "CyclogeostrophyResult",
    "CyclogeostrophySetup",
]



import jax
import jax.numpy as jnp
from jaxtyping import Float

from .geometry import coriolis_factor, grid_metrics
from .operators import horizontal_derivatives, interpolation
from .sanitize import init_land_mask, sanitize_data


def setup_kinematics(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
):
    land_mask = init_land_mask(u, land_mask)

    if not uv_on_t:
        u = interpolation(u, axis=1, padding="left", land_mask=land_mask)  # U(i), U(i+1) -> T(i+1)
        v = interpolation(v, axis=0, padding="left", land_mask=land_mask)  # V(j), V(j+1) -> T(j+1)

    if lat_t is None or lon_t is None:
        if lat_u is not None and lon_u is not None:
            lat_t = interpolation(lat_u, axis=1, padding="left", land_mask=land_mask)
            lon_t = interpolation(lon_u, axis=1, padding="left", land_mask=land_mask)
        elif lat_v is not None and lon_v is not None:
            lat_t = interpolation(lat_v, axis=0, padding="left", land_mask=land_mask)
            lon_t = interpolation(lon_v, axis=0, padding="left", land_mask=land_mask)
        else:
            raise ValueError("Either lat_t and lon_t, or lat_u, lon_u, lat_v, and lon_v must be provided")
    
    # compute grid metrics once
    dx_e, dx_n, dy_e, dy_n, J = grid_metrics(lat_t, lon_t)

    return u, v, lat_t, lon_t, dx_e, dx_n, dy_e, dy_n, J, land_mask


def magnitude(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True
) -> Float[jax.Array, "y x"]:
    """
    Computes the magnitude (azimuthal velocity) of a 2d velocity field.

    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    Parameters
    ----------
    u : Float[jax.Array, "y x"]
        $u$ component of the velocity field (on the U or T grid)
    v : Float[jax.Array, "y x"]
        $v$ component of the velocity field (on the V or T grid)
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

        Defaults to `None`
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the U and V grids,
        and are interpolated to the T one (following NEMO convention).
        If `False`, the velocity components are assumed to be located on the T grid, and interpolation is not needed.
        
        Defaults to `True`

    Returns
    -------
    magnitude : Float[jax.Array, "y x"]
        Magnitude of the velocity field, on the T grid
    """
    land_mask = init_land_mask(u, land_mask)

    if not uv_on_t:
        u = interpolation(u, axis=1, padding="left", land_mask=land_mask)  # U(i), U(i+1) -> T(i+1)
        v = interpolation(v, axis=0, padding="left", land_mask=land_mask)  # V(j), V(j+1) -> T(j+1)

    magn = jnp.sqrt(u** 2 + v ** 2)
    magn = sanitize_data(magn, jnp.nan, land_mask)

    return magn


def kinetic_energy(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
) -> Float[jax.Array, "y x"]:
    """
    Computes the Kinetic Energy (KE) of a velocity field.
    
    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    Parameters
    ----------
    u : Float[jax.Array, "y x"]
        $u$ component of the velocity field (on the U grid)
    v : Float[jax.Array, "y x"]
        $v$ component of the velocity field (on the V grid)
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land).

        Defaults to `None`
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the U and V grids,
        and are interpolated to the T one (following NEMO convention).
        If `False`, the velocity components are assumed to be located on the T grid, and interpolation is not needed.

        Defaults to `True`

    Returns
    -------
    kinetic_energy : Float[jax.Array, "y x"]
        The Kinetic Energy on the T grid
    """
    # Make sure the mask is initialized
    land_mask = init_land_mask(u, land_mask)
    
    if not uv_on_t:
        u = interpolation(u, axis=1, padding="left", land_mask=land_mask)  # U(i), U(i+1) -> T(i+1)
        v = interpolation(v, axis=0, padding="left", land_mask=land_mask)  # V(j), V(j+1) -> T(j+1)

    ke = (u ** 2 + v ** 2) / 2
    ke = sanitize_data(ke, jnp.nan, land_mask)

    return ke


def vorticity(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
    normalize_by_coriolis: bool = True
) -> Float[jax.Array, "y x"]:
    """
    Computes the relative vorticity of a velocity field.

    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    If provided, the ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are expected to follow the NEMO convention.

    Parameters
    ----------
    u : Float[jax.Array, "y x"]
        $u$ component of the velocity field
    v : Float[jax.Array, "y x"]
        $v$ component of the velocity field
    lat_t : Float[jax.Array, "y x"], optional
        Latitudes of the T grid.
        
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lon_t : Float[jax.Array, "y x"], optional
        Longitudes of the T grid.
       
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lat_u : Float[jax.Array, "y x"], optional
        Latitudes of the U grid.
        
        Defaults to `None`
    lon_u : Float[jax.Array, "y x"], optional
        Longitudes of the U grid.
        
        Defaults to `None`
    lat_v : Float[jax.Array, "y x"], optional
        Latitudes of the V grid.
        
        Defaults to `None`
    lon_v : Float[jax.Array, "y x"], optional
        Longitudes of the V grid.
        
        Defaults to `None`
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the T grid 
        (this is important when manipulating staggered grids)
        
        Defaults to `True`
    normalize_by_coriolis : bool, optional
        If `True`, returns the vorticity normalized by the Coriolis factor

        Defaults to `True`

    Returns
    -------
    vorticity : Float[jax.Array, "y x"]
        The vorticity on the T grid, normalized by the Coriolis factor if ``normalize_by_coriolis=True``
    """
    u, v, lat_t, lon_t, dx_e, dx_n, dy_e, dy_n, J, land_mask = setup_kinematics(
        u, v, lat_t, lon_t, lat_u, lon_u, lat_v, lon_v, land_mask, uv_on_t
    )

    _, du_n = horizontal_derivatives(u, dx_e=dx_e, dx_n=dx_n, dy_e=dy_e, dy_n=dy_n, J=J, land_mask=land_mask)
    dv_e, _ = horizontal_derivatives(v, dx_e=dx_e, dx_n=dx_n, dy_e=dy_e, dy_n=dy_n, J=J, land_mask=land_mask)

    vort = dv_e - du_n

    if normalize_by_coriolis:
        f = coriolis_factor(lat_t)
        vort /= f

    vort = sanitize_data(vort, jnp.nan, land_mask)

    return vort


def strain_rate(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
    normalize_by_coriolis: bool = True
) -> Float[jax.Array, "y x"]:
    """
    Computes the strain rate magnitude of a velocity field.

    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    If provided, the ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are expected to follow the NEMO convention.

    Parameters
    ----------
    u : Float[jax.Array, "y x"]
        $u$ component of the velocity field
    v : Float[jax.Array, "y x"]
        $v$ component of the velocity field
    lat_t : Float[jax.Array, "y x"], optional
        Latitudes of the T grid.
        Defaults to `None`
    lon_t : Float[jax.Array, "y x"], optional
        Longitudes of the T grid.
        Defaults to `None`
    lat_u : Float[jax.Array, "y x"], optional
        Latitudes of the U grid.
        Defaults to `None`
    lon_u : Float[jax.Array, "y x"], optional
        Longitudes of the U grid.
        Defaults to `None`
    lat_v : Float[jax.Array, "y x"], optional
        Latitudes of the V grid.
        Defaults to `None`
    lon_v : Float[jax.Array, "y x"], optional
        Longitudes of the V grid.
        Defaults to `None`
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the T grid 
        (this is important when manipulating staggered grids)
        
        Defaults to `True`
    normalize_by_coriolis : bool, optional
        If `True`, returns the strain rate normalized by the Coriolis factor

        Defaults to `True`

    Returns
    -------
    strain_rate : Float[jax.Array, "y x"]
        The strain rate magnitude on the T grid, normalized by the Coriolis factor if ``normalize_by_coriolis=True``
    """
    u, v, lat_t, lon_t, dx_e, dx_n, dy_e, dy_n, J, land_mask = setup_kinematics(
        u, v, lat_t, lon_t, lat_u, lon_u, lat_v, lon_v, land_mask, uv_on_t
    )

    du_e, du_n = horizontal_derivatives(u, dx_e=dx_e, dx_n=dx_n, dy_e=dy_e, dy_n=dy_n, J=J, land_mask=land_mask)
    dv_e, dv_n = horizontal_derivatives(v, dx_e=dx_e, dx_n=dx_n, dy_e=dy_e, dy_n=dy_n, J=J, land_mask=land_mask)

    strain = jnp.sqrt((du_e - dv_n) ** 2 + (dv_e + du_n) ** 2)

    if normalize_by_coriolis:
        f = coriolis_factor(lat_t)
        strain /= f

    strain = sanitize_data(strain, jnp.nan, land_mask)

    return strain


def radius_of_curvature(
    u: Float[jax.Array, "y x"],
    v: Float[jax.Array, "y x"],
    lat_t: Float[jax.Array, "y x"] = None,
    lon_t: Float[jax.Array, "y x"] = None,
    lat_u: Float[jax.Array, "y x"] = None,
    lon_u: Float[jax.Array, "y x"] = None,
    lat_v: Float[jax.Array, "y x"] = None,
    lon_v: Float[jax.Array, "y x"] = None,
    land_mask: Float[jax.Array, "y x"] = None,
    uv_on_t: bool = True,
) -> Float[jax.Array, "y x"]:
    """
    Computes the radius of curvature of a 2d velocity field.

    The velocity field can be provided either on the T grid (``uv_on_t=True``) or on the U/V grids (``uv_on_t=False``).

    If provided, the ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are expected to follow the NEMO convention.

    Parameters
    ----------
    u : Float[jax.Array, "y x"]
        $u$ component of the velocity field
    v : Float[jax.Array, "y x"]
        $v$ component of the velocity field
    lat_t : Float[jax.Array, "y x"], optional
        Latitudes of the T grid.
        
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lon_t : Float[jax.Array, "y x"], optional
        Longitudes of the T grid.
       
        If ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are not provided, ``lat_t`` and ``lon_t`` must be provided to compute them.
        
        Defaults to `None`
    lat_u : Float[jax.Array, "y x"], optional
        Latitudes of the U grid.
        
        Defaults to `None`
    lon_u : Float[jax.Array, "y x"], optional
        Longitudes of the U grid.
        
        Defaults to `None`
    lat_v : Float[jax.Array, "y x"], optional
        Latitudes of the V grid.
        
        Defaults to `None`
    lon_v : Float[jax.Array, "y x"], optional
        Longitudes of the V grid.
        
        Defaults to `None`
    land_mask : Float[jax.Array, "y x"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    uv_on_t : bool, optional
        If `True`, the velocity components are assumed to be located on the T grid 
        (this is important when manipulating staggered grids)
        
        Defaults to `True`

    Returns
    -------
    rc : Float[jax.Array, "y x"]
        The radius of curvature of the velocity field in meters, on the T grid
    """
    u, v, lat_t, lon_t, dx_e, dx_n, dy_e, dy_n, J, land_mask = setup_kinematics(
        u, v, lat_t, lon_t, lat_u, lon_u, lat_v, lon_v, land_mask, uv_on_t
    )

    return _radius_of_curvature(u, v, dx_e, dx_n, dy_e, dy_n, J, land_mask)


def _radius_of_curvature(
    u_t: Float[jax.Array, "y x"],
    v_t: Float[jax.Array, "y x"],
    dx_e_t: Float[jax.Array, "y x"],
    dx_n_t: Float[jax.Array, "y x"],
    dy_e_t: Float[jax.Array, "y x"],
    dy_n_t: Float[jax.Array, "y x"],
    J_t: Float[jax.Array, "y x"],
    land_mask: Float[jax.Array, "y x"]
) -> Float[jax.Array, "y x"]:
    V_t = magnitude(u_t, v_t, land_mask, uv_on_t=True)

    du_e_t, du_n_t = horizontal_derivatives(
        u_t, dx_e=dx_e_t, dx_n=dx_n_t, dy_e=dy_e_t, dy_n=dy_n_t, J=J_t, land_mask=land_mask
    )
    dv_e_t, dv_n_t = horizontal_derivatives(
        v_t, dx_e=dx_e_t, dx_n=dx_n_t, dy_e=dy_e_t, dy_n=dy_n_t, J=J_t, land_mask=land_mask
    )

    numerator = V_t ** 3
    denominator = u_t ** 2 * dv_e_t - v_t ** 2 * du_n_t - u_t * v_t * (du_e_t - dv_n_t)
    r = numerator / denominator

    return r
